#include "dumpwtb.h"
#include "flagbuilder.h"
#include "disassembler.h"
#include "labelmap.h"
#include "welltype/version.h"
#include "welltype/wtruntime.h"
#include "welltype/byteorder.h"
#include "wtruntime/hash.h"
#include "libwtref/libwtref.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct dumpwtb_state
{
	const wt_u8*         data8;
	size_t               datalen;
	const wtb_hdr*       hdr;
	struct FlagBuilder*  sec_flags;
	struct FlagBuilder*  thr_flags;
	struct Disassembler* disasm;
	struct labelmap*     labels;
	struct labelmap*     used_labels;
	const char**         type_array;
	size_t               type_count;
	wt_bool              vermatch;
	wt_u32               bin_version;
	wt_bo                bo;
};

#define PUT_NIBBLE(_buffer_,_nibble_) \
	do \
	{ \
		if((_nibble_)>9) \
		{ \
			*(_buffer_)++ = 'A' + ((_nibble_) - 10); \
		} \
		else \
		{ \
			*(_buffer_)++ = '0' + (_nibble_); \
		} \
	} \
	while(0)

static const char* g_byteorder_names[] =
{
	"?",
	"little-endian",
	"big-endian"
};

static const char* g_bos_fields[] =
{
	"Octets",
	"16-bit",
	"32-bit",
	"64-bit"
};

static const char* g_assoc_names[] =
{
	"?",
	"left",
	"right"
};

static const char* render64_u(wt_u64 n)
{
	static char buffer[32];
	char* ptr = buffer + sizeof(buffer);
	*--ptr = 0;

	do
	{
		*--ptr = '0' + (unsigned int)(n % 10);
		n /= 10;
	}
	while(0!=n);

	return ptr;
}

static const char* render64_016X(wt_u64 n)
{
	static char buffer[32];
	int i;

	for(i=15;0<=i;--i)
	{
		const unsigned int nibble = (unsigned int)(n & 0xf);
		n >>= 4;

		if(nibble<10)
		{
			buffer[i] = '0' + nibble;
		}
		else
		{
			buffer[i] = 'A' + (nibble - 10);
		}
	}

	return buffer;
}

static const char* section_type(wt_u8 type, char* buf, wt_u32 size)
{
	wt_u8 hi  = type >> 4;
	wt_u8 lo  = type & 0xf;
	char* ret = buf;

	if(size<8) { return "?NOFIT?"; }

	switch(type)
	{
	case WTBST_UNUSED:     return "UNUSED";
	case WTBST_CODE:       return "CODE";
	case WTBST_DATA:       return "DATA";
	case WTBST_BOS:        return "BOS";
	case WTBST_IMPORT:     return "IMPORT";
	case WTBST_EXPORT:     return "EXPORT";
	case WTBST_METADATA:   return "METADATA";
	case WTBST_INTEGRITY:  return "INTEGRITY";
	case WTBST_DEBUGLABEL: return "DEBUGLABEL";
	}

	if(WTBST_CUSTOMLO<=type && type<=WTBST_CUSTOMHI)
	{
		*buf++ = 'C';
		*buf++ = 'U';
		*buf++ = 'S';
	}
	else
	{
		*buf++ = 'U';
		*buf++ = 'N';
		*buf++ = 'K';
	}

	*buf++ = '_';
	*buf++ = '0';
	*buf++ = 'x';

	PUT_NIBBLE(buf, hi);
	PUT_NIBBLE(buf, lo);

	*buf++ = '\0';

	return ret;
}

static const char* thread_type(wt_u16 type, char* buf, wt_u32 size)
{
	wt_u8 hihi = type >> 12;
	wt_u8 hilo = (type >> 8) & 0xf;
	wt_u8 lohi = (type >> 4) & 0xf;
	wt_u8 lolo = type & 0xf;
	char* ret  = buf;

	if(size<8) { return "?NOFIT?"; }

	switch(type)
	{
	case WTBTT_USER: return "USER";
	case WTBTT_TASK: return "TASK";
	}

	*buf++ = 'U';
	*buf++ = 'N';
	*buf++ = 'K';
	*buf++ = '_';

	PUT_NIBBLE(buf, hihi);
	PUT_NIBBLE(buf, hilo);
	PUT_NIBBLE(buf, lohi);
	PUT_NIBBLE(buf, lolo);

	*buf++ = '\0';

	return ret;
}

static int dumpwtb_bos(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wt_u8* bos, wt_u32 size)
{
	const wt_u8* end = bos + size;

	printf("Byteorder Swap section at offset 0x%X contains:\n", offs);

	printf("  Addr     Field  Count\n");

	while(bos<end)
	{
		wt_u64 count = 0;
		wt_u32 addr0 = addr;
		wt_u8  fsize = (*bos)>>4;
		wt_u8  csize = (*bos)&0xf;

		if(fsize>0x3 || csize>0x4)
		{
			printf("Invalid BOS entry.\n");
			return 1;
		}

		++bos;

		if(0==csize)
		{
			continue;
		}

		switch(csize)
		{
		case 1:
			{
				count = *bos++;
			}
			break;

		case 2:
			{
				wt_u16 w;
				memcpy(&w, bos, sizeof(w));
				count = wtr_botoh16(w, st->bo);
				bos += sizeof(w);
			}
			break;

		case 3:
			{
				wt_u32 w;
				memcpy(&w, bos, sizeof(w));
				count = wtr_botoh32(w, st->bo);
				bos += sizeof(w);
			}
			break;

		case 4:
			{
				wt_u64 w;
				memcpy(&w, bos, sizeof(w));
				count = wtr_botoh64(w, st->bo);
				bos += sizeof(w);
			}
			break;
		}

		addr += (wt_u32)(count << fsize);
		printf("  %08X %-5s %s\n", addr0, g_bos_fields[fsize], render64_u(count));
	}

	printf("\n");

	return 0;
}

#define NAME_WIDTH "30"

static void dumpwtb_show_type(struct dumpwtb_state* const st, wt_u32 type);

static void dumpwtb_show_type(struct dumpwtb_state* const st, wt_u32 type)
{
	if(type & WTTF_REF)
	{
		type &= ~WTTF_REF;
		printf("ref ");
	}

	if(type & WTTF_MUTABLE)
	{
		type &= ~WTTF_MUTABLE;
		printf("mutable ");
	}

	if(type & WTTF_SEQ)
	{
		type &= ~WTTF_SEQ;
		printf("seq<");
		dumpwtb_show_type(st, type);
		printf(">");
	}
	else
	if(type & WTTF_LOCAL)
	{
		const char** const p_name = st->type_array + (type & WTTM_USERTYPEID);

		if(NULL!=p_name)
		{
			printf("%s", *p_name);
		}
		else
		{
			printf("?");
		}
	}
	else
	{
		switch(type)
		{
		case WTT_VOID:    printf("void");       return;
		case WTT_BOOL:    printf("bool");       return;
		case WTT_I8:      printf("i8");         return;
		case WTT_U8:      printf("u8");         return;
		case WTT_I16:     printf("i16");        return;
		case WTT_U16:     printf("u16");        return;
		case WTT_I32:     printf("i32");        return;
		case WTT_U32:     printf("u32");        return;
		case WTT_I64:     printf("i64");        return;
		case WTT_U64:     printf("u64");        return;
		case WTT_F32:     printf("f32");        return;
		case WTT_F64:     printf("f64");        return;
		case WTT_CHAR:    printf("char");       return;
		case WTT_STRING:  printf("string");     return;
		case WTT_HANDLE:  printf("handle");     return;
		case WTT_FILE:    printf("file");       return;
		case WTT_DIR:     printf("dir");        return;
		case WTT_PROGRAM: printf("wt_program"); return;
		}

		printf("?Unk?_%08X", type);
	}
}

static void dumpwtb_register_type(struct dumpwtb_state* const st, wt_u32 idx, const char* name)
{
	if(st->type_count<=idx)
	{
		if(NULL==(st->type_array=(const char**)realloc(st->type_array, sizeof(const char*)*(idx+1))))
		{
			abort();
		}

		memset(st->type_array+st->type_count, 0, sizeof(const char*)*(idx+1-st->type_count));
		st->type_count = idx + 1;
	}

	st->type_array[idx] = name;
}

static void dumpwtb_show_funcsig(struct dumpwtb_state* const st, const wt_u32* ptr)
{
	wt_u32 cnt;
	wt_u32 i;

	printf("(");

	for(i=0,cnt=wtr_botoh32(*ptr++, st->bo);cnt>0;++i,--cnt)
	{
		if(0<i)
		{
			printf(", ");
		}

		dumpwtb_show_type(st, wtr_botoh32(*ptr++, st->bo));
	}

	printf(")");

	for(i=0,cnt=wtr_botoh32(*ptr++, st->bo);cnt>0;++i,--cnt)
	{
		if(0==i)
		{
			printf(" : ");
		}
		else
		{
			printf(", ");
		}

		dumpwtb_show_type(st, wtr_botoh32(*ptr++, st->bo));
	}

	{
		const wt_u8* p_pure = (const wt_u8*)ptr;

		if(*p_pure)
		{
			printf(" [pure]");
		}
		else
		{
			printf(" [impure]");
		}
	}
}

static void dumpwtb_show_record(struct dumpwtb_state* const st, const wt_u32* ptr)
{
#define PREFIX "             "

	const char* names;
	wt_u32      cnt;
	wt_u32      i;

	printf(PREFIX "{\n");

	for(i=0,cnt=wtr_botoh32(*ptr++, st->bo),names=(const char*)(ptr+cnt);cnt>0;++i,--cnt)
	{
		printf(PREFIX "  ");
		dumpwtb_show_type(st, wtr_botoh32(*ptr++, st->bo));
		printf(" %s;\n", names);
		names += strlen(names) + 1;
	}

	printf(PREFIX "}\n");

#undef PREFIX
}

static void dumpwtb_show_datatype(struct dumpwtb_state* const st, const void* ptr)
{
#define PREFIX "             "

	wt_u32 cons_num = wtr_botoh32(((const wt_u32*)ptr)[0], st->bo);
	wt_u32 def_cons = wtr_botoh32(((const wt_u32*)ptr)[1], st->bo);
	wt_u32 i;

	ptr = ((const char*)ptr) + sizeof(wt_u32)*2;

	for(i=0;i<cons_num;++i)
	{
		const char*   name = (const char*)ptr;
		const wt_u32  nlen = strlen(name);
		const wt_u32* us = (const wt_u32*)(name + ((nlen + sizeof(wt_u32)) & ((~sizeof(wt_u32))+1)));
		wt_u32        typn = wtr_botoh32(*us++, st->bo);
		wt_u32        j;

		printf(PREFIX "#%u %s%s", (i+1), ((i+1)==def_cons)?"default ":"", name);

		for(j=0;j<typn;++j)
		{
			if(j)
			{
				printf(", ");
			}
			else
			{
				printf(": ");
			}

			dumpwtb_show_type(st, wtr_botoh32(*us++, st->bo));
		}

		printf("\n");

		ptr = us;
	}

#undef PREFIX
}

static void dumpwtb_show_enum(struct dumpwtb_state* const st, const void* ptr)
{
#define PREFIX "             "

	wt_u32 enumc = wtr_botoh32(*((const wt_u32*)ptr), st->bo);
	wt_u32 i;

	ptr = ((const char*)ptr) + sizeof(wt_u32);

	for(i=0;i<enumc;++i)
	{
		const char* name = (const char*)ptr;

		printf(PREFIX "#%u %s\n", (i+1), name);

		ptr = ((const char*)ptr) + strlen(name) + 1;
	}

#undef PREFIX
}

static int dumpwtb_import(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wtb_symhdr* symhdr, wt_u32 size)
{
	const wtb_sym* sym      = (const wtb_sym*)(symhdr + 1);
	wt_u32         symcount = 0;

	if(size<sizeof(wtb_symhdr))
	{
		printf("Import section size is less than expected.\n");
		return 1;
	}

	symcount = wtr_botoh32(symhdr->symcount, st->bo);
	printf("Import section at offset 0x%X contains %u entries:\n", offs, symcount);

	printf("  Type       Info     %-" NAME_WIDTH "s Data\n", "Name");

	for(;symcount;++sym,--symcount)
	{
		wt_u32      sym_type = wtr_botoh32(sym->sym_type, st->bo);
		wt_u32      sym_name = wtr_botoh32(sym->sym_name, st->bo);
		wt_u32      sym_offs = wtr_botoh32(sym->sym_offs, st->bo);
		wt_u32      sym_info = wtr_botoh32(sym->sym_info, st->bo);
		const char* name     = ((const char*)symhdr) + sym_name;

		switch(sym_type)
		{
		case WTBSYMT_RECORD:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  record     %08X %-" NAME_WIDTH "." NAME_WIDTH "s\n", sym_info, name);
				dumpwtb_show_record(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		case WTBSYMT_PRIVATERECORD:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  record     %08X %-" NAME_WIDTH "." NAME_WIDTH "s private\n", sym_info, name);
			}
			break;

		case WTBSYMT_LIMITEDRECORD:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  record     %08X %-" NAME_WIDTH "." NAME_WIDTH "s limited\n", sym_info, name);
			}
			break;

		case WTBSYMT_DATATYPE:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  datatype   %08X %-" NAME_WIDTH "." NAME_WIDTH "s\n", sym_info, name);
				dumpwtb_show_datatype(st, (((char*)symhdr) + sym_offs));
			}
			break;

		case WTBSYMT_ENUM:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  enum       %08X %-" NAME_WIDTH "." NAME_WIDTH "s\n", sym_info, name);
				dumpwtb_show_enum(st, (((char*)symhdr) + sym_offs));
			}
			break;

		case WTBSYMT_EXCEPTION:
			{
				printf("  exception  %08X %-" NAME_WIDTH "." NAME_WIDTH "s -\n", sym_info, name);
			}
			break;

		case WTBSYMT_OPERATOR:
			{
				wt_u32 assoc = WTBSYM_OPASSOC(sym_info);
				const char* assoc_str = g_assoc_names[assoc];
				if(WTBSYM_OP_LEFTASSOC!=assoc && WTBSYM_OP_RIGHTASSOC!=assoc) { assoc_str = "?"; }
				printf("  operator   %08X %-" NAME_WIDTH "." NAME_WIDTH "s %s %u\n", sym_info, name, assoc_str, WTBSYM_OPPREC(sym_info));
			}
			break;

		case WTBSYMT_FUNCTION:
			{
				printf("  function   %08X %-" NAME_WIDTH "." NAME_WIDTH "s ", sym_info, name);
				dumpwtb_show_funcsig(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		case WTBSYMT_FUNCTIONTYPE:
			{
				printf("  func_type  %08X %-" NAME_WIDTH "." NAME_WIDTH "s ", sym_info, name);
				dumpwtb_show_funcsig(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		default:
			{
				printf("  unknown <0x%08X>\n", sym_type);
			}
			break;
		}
	}

	printf("\n");

	return 0;
}

static int dumpwtb_export(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wtb_symhdr* symhdr, wt_u32 size)
{
	const wtb_sym* sym      = (const wtb_sym*)(symhdr + 1);
	wt_u32         symcount = 0;

	if(size<sizeof(wtb_symhdr))
	{
		printf("Export section size is less than expected.\n");
		return 1;
	}

	symcount = wtr_botoh32(symhdr->symcount, st->bo);
	printf("Export section at offset 0x%X contains %u entries:\n", offs, symcount);

	printf("  Type       Info     %-" NAME_WIDTH "s Data\n", "Name");

	for(;symcount;++sym,--symcount)
	{
		wt_u32      sym_type = wtr_botoh32(sym->sym_type, st->bo);
		wt_u32      sym_name = wtr_botoh32(sym->sym_name, st->bo);
		wt_u32      sym_offs = wtr_botoh32(sym->sym_offs, st->bo);
		wt_u32      sym_info = wtr_botoh32(sym->sym_info, st->bo);
		const char* name     = ((const char*)symhdr) + sym_name;

		switch(sym_type)
		{
		case WTBSYMT_RECORD:
			{
				printf("  record     -        %-" NAME_WIDTH "." NAME_WIDTH "s\n", name);
				dumpwtb_show_record(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		case WTBSYMT_PRIVATERECORD:
			{
				printf("  record     -        %-" NAME_WIDTH "." NAME_WIDTH "s private\n", name);
			}
			break;

		case WTBSYMT_LIMITEDRECORD:
			{
				printf("  record     -        %-" NAME_WIDTH "." NAME_WIDTH "s limited\n", name);
			}
			break;

		case WTBSYMT_DATATYPE:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  datatype   %08X %-" NAME_WIDTH "." NAME_WIDTH "s\n", sym_info, name);
				dumpwtb_show_datatype(st, (((char*)symhdr) + sym_offs));
			}
			break;

		case WTBSYMT_ENUM:
			{
				dumpwtb_register_type(st, sym_info, name);
				printf("  enum       %08X %-" NAME_WIDTH "." NAME_WIDTH "s\n", sym_info, name);
				dumpwtb_show_enum(st, (((char*)symhdr) + sym_offs));
			}
			break;

		case WTBSYMT_EXCEPTION:
			{
				printf("  exception  -        %-" NAME_WIDTH "." NAME_WIDTH "s -\n", name);
			}
			break;

		case WTBSYMT_OPERATOR:
			{
				wt_u32 assoc = WTBSYM_OPASSOC(sym_info);
				const char* assoc_str = g_assoc_names[assoc];
				if(WTBSYM_OP_LEFTASSOC!=assoc && WTBSYM_OP_RIGHTASSOC!=assoc) { assoc_str = "?"; }
				printf("  operator   %08X %-" NAME_WIDTH "." NAME_WIDTH "s %s %u\n", sym_info, name, assoc_str, WTBSYM_OPPREC(sym_info));
			}
			break;

		case WTBSYMT_FUNCTION:
			{
				printf("  function   %08X %-" NAME_WIDTH "." NAME_WIDTH "s ", sym_info, name);
				dumpwtb_show_funcsig(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		case WTBSYMT_FUNCTIONTYPE:
			{
				printf("  func_type  %08X %-" NAME_WIDTH "." NAME_WIDTH "s ", sym_info, name);
				dumpwtb_show_funcsig(st, (wt_u32*)(((char*)symhdr) + sym_offs));
				printf("\n");
			}
			break;

		default:
			{
				printf("  unknown <0x%08X>\n", sym_type);
			}
			break;
		}
	}

	printf("\n");

	return 0;
}

static void dumpwtb_printcol(const wt_u8* data, const wt_u8* end, wt_u32 col_width, wt_bool hex)
{
	wt_u32      index;
	const char* format = hex ? " %02X" : "%c";
	const char* space  = hex ? "   " : "";

	for(index=0;index<col_width;++index)
	{
		if(data+index<end)
		{
			wt_u8 ch = data[index];

			if(!hex && (ch<0x20 || ch>=0x7F))
			{
				ch = '.';
			}

			printf(format, ch);
		}
		else
		{
			printf("%s", space);
		}
	}
}

static int dumpwtb_data(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wt_u8* data, wt_u32 data_size, const wt_u8* bos, wt_u32 bos_size, wt_bool assume_str, wt_bool flag_clear)
{
	const wt_u8* const end = data + data_size;

	printf("Data section at offset 0x%X contains:\n", offs);

	if(bos)
	{
		const wt_u8* bos_end = bos + bos_size;

		printf("  Addr     Field  Value\n");

		while(data<end && bos<bos_end)
		{
			wt_u64 count = 0;
			wt_u8  fsize = (*bos)>>4;
			wt_u8  csize = (*bos)&0xf;

			const char* field_name;

			if(fsize>0x3 || csize==0x0 || csize>0x4)
			{
				printf("Invalid BOS entry.\n");
				return 1;
			}

			field_name = g_bos_fields[fsize];
			++bos;

			switch(csize)
			{
			case 1:
				{
					count = *bos++;
				}
				break;

			case 2:
				{
					wt_u16 w;
					memcpy(&w, bos, sizeof(w));
					count = wtr_botoh16(w, st->bo);
					bos += sizeof(w);
				}
				break;

			case 3:
				{
					wt_u32 w;
					memcpy(&w, bos, sizeof(w));
					count = wtr_botoh32(w, st->bo);
					bos += sizeof(w);
				}
				break;

			case 4:
				{
					wt_u64 w;
					memcpy(&w, bos, sizeof(w));
					count = wtr_botoh64(w, st->bo);
					bos += sizeof(w);
				}
				break;
			}

			switch(fsize)
			{
			case 0:
				if(assume_str)
				{
					const wt_u8* const fend  = data + count;
					wt_u32             width = 32;

					const char* str_name = "String";

					while(data<fend)
					{
						wt_u32  line_width = 0;
						wt_bool zero       = ('\0'==data[line_width]);
						wt_u32  max_width  = zero ? (width / 2) : width;

						while(line_width<max_width && (data+line_width)<fend && ((zero && !data[line_width]) || (!zero && data[line_width])))
						{
							++line_width;
						}

						if(!zero && line_width<max_width) { ++line_width; }

						printf("  %08X %-5s%s", addr, zero ? field_name : str_name, zero?"":" \"");
						dumpwtb_printcol(data, fend, line_width, zero);
						printf("%s\n", zero?"":"\"");

						addr += line_width;
						data += line_width;
					}
				}
				else
				{
					const wt_u8* const fend  = data + count;
					wt_u32             width = 16;

					while(data<fend)
					{
						wt_u32 line_width = width;

						if((size_t)(fend-data) < (size_t)line_width)
						{
							line_width = (wt_u32)(fend-data);
						}

						printf("  %08X %-5s", addr, field_name);
						dumpwtb_printcol(data, fend, line_width, WT_TRUE);
						printf("\n");
						addr += line_width;
						data += line_width;
					}
				}
				break;

			case 1:
				{
					const wt_u16*       field = (const wt_u16*)data;
					const wt_u16* const fend  = field + count;

					while(field<fend)
					{
						printf("  %08X %-5s %04X\n", addr, field_name, wtr_botoh16(*field, st->bo));
						addr += sizeof(*field);
						data += sizeof(*field);
						++field;
					}
				}
				break;

			case 2:
				{
					const wt_u32*       field = (const wt_u32*)data;
					const wt_u32* const fend  = field + count;

					while(field<fend)
					{
						printf("  %08X %-5s %08X\n", addr, field_name, (unsigned int)wtr_botoh32(*field, st->bo));
						addr += sizeof(*field);
						data += sizeof(*field);
						++field;
					}
				}
				break;

			case 3:
				{
					const wt_u64*       field = (const wt_u64*)data;
					const wt_u64* const fend  = field + count;

					while(field<fend)
					{
						printf("  %08X %-5s %s\n", addr, field_name, render64_016X(wtr_botoh64(*field, st->bo)));
						addr += sizeof(*field);
						data += sizeof(*field);
						++field;
					}
				}
				break;
			}
		}

		printf("\n");
	}
	else
	if(flag_clear)
	{
		wt_u32 col_width = 8;
		wt_u8  empty[8]  = { 0, 0, 0, 0, 0, 0, 0, 0 };

		while(data<end)
		{
			printf("  %08X:", addr);

			dumpwtb_printcol(empty, empty+(end-data), col_width, WT_TRUE);
			printf(" ");
			dumpwtb_printcol(empty, empty+(end-(data+8)), col_width, WT_TRUE);
			printf(" |");
			dumpwtb_printcol(empty, empty+(end-data), col_width*2, WT_FALSE);
			printf("|\n");

			addr += col_width*2;
			data += col_width*2;
		}

		printf("\n");
	}
	else
	{
		wt_u32 col_width = 8;

		while(data<end)
		{
			printf("  %08X:", addr);

			dumpwtb_printcol(data, end, col_width, WT_TRUE);
			printf(" ");
			dumpwtb_printcol(data+8, end, col_width, WT_TRUE);
			printf(" |");
			dumpwtb_printcol(data, end, col_width*2, WT_FALSE);
			printf("|\n");

			addr += col_width*2;
			data += col_width*2;
		}

		printf("\n");
	}

	return 0;
}

static int dumpwtb_code(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wt_u16* code, wt_u32 size)
{
	const wt_u16* result;

	printf("Disassembly of code section at offset 0x%X:\n", offs);

	if((result=Disassembler_Disassemble(st->disasm, st->vermatch, addr, code, size, st->used_labels, st->bo))!=NULL)
	{
		printf("Undefined instruction at offset 0x%X. Instruction: 0x%04X.\n", offs + (wt_u32)(result - code), wtr_botoh16(*result, st->bo));
		return 1;
	}

	printf("\n");

	return 0;
}

static int dumpwtb_integrity(struct dumpwtb_state* const st, wt_u32 addr, wt_u32 offs, const wtb_integhdr* ihdr, wt_u32 size)
{
	wt_u16       alg_id  = wtr_botoh16(ihdr->algorithm, st->bo);
	wt_u16       diglen  = 0;
	size_t       width   = 0;
	const wt_u8* digest  = (const wt_u8*)(ihdr + 1);
	const char*  alg_str = NULL;
	const char*  result  = NULL;

	const struct wtr_hashalg* const alg = wtr_hash_find_by_id(
		libwtref_hash_algs, alg_id);

	if(size<sizeof(wtb_integhdr))
	{
		printf("Integrity section size is less than expected.\n");
		return 1;
	}

	if(0!=ihdr->reserved)
	{
		printf("Reserved field is not zero.\n");
		return 1;
	}

	if(NULL!=alg)
	{
		if(0!=alg->digest_len && alg->digest_len+sizeof(wtb_integhdr)!=size)
		{
			printf("Unexpected section size.\n");
			return 1;
		}

		diglen  = alg->digest_len;
		alg_str = alg->name;

		if(WTE_OK==wtr_hash_check(ihdr+1, alg, st->data8, st->datalen, st->bo))
		{
			result = "verified, OK";
		}
		else
		{
			result = "cannot be verified";
		}
	}
	else
	{
		alg_str = "<unknown>";
		result = "hash algorithm not supported";
	}

	printf("Integrity section at offset 0x%X contains:\n", offs);
	printf("  Integrity: %s\n", result);
	printf("  Algorithm: %s\n", alg_str);
	printf("  Digest:   ");
	while(0<(diglen--))
	{
		printf(" %02X", *digest++);

		if((++width)>=16 && diglen)
		{
			printf("\n");
			printf("            ");

			width = 0;
		}
	}
	printf("\n");
	printf("\n");

	return 0;
}

static int dumpwtb_metadata(struct dumpwtb_state* const st, wt_u32 offs, const wtb_metahdr* metahdr, wt_u32 size)
{
	const unsigned char* last      = ((const unsigned char*)metahdr) + size - 1;
	const wtb_meta*      meta      = (const wtb_meta*)(metahdr + 1);
	wt_u32               metacount = 0; 

	if(size<sizeof(wtb_metahdr))
	{
		printf("Metadata section size is less than expected.\n");
		return 1;
	}

	if(*last)
	{
		printf("Last octet of the metadata section is not zero.\n");
		return 1;
	}

	metacount = wtr_botoh32(metahdr->metacount, st->bo);
	printf("Metadata table at offset 0x%X contains %u entries:\n", offs, metacount);

	for(;metacount;++meta,--metacount)
	{
		wt_u32 key_offs = wtr_botoh32(meta->meta_key_offs, st->bo);
		wt_u32 val_offs = wtr_botoh32(meta->meta_val_offs, st->bo);

		if(key_offs>=size || val_offs>=size)
		{
			printf("Metadata key/value offset is out of range.\n");
			return 1;
		}
		else
		{
			const char* key = ((const char*)metahdr) + key_offs;
			const char* val = ((const char*)metahdr) + val_offs;

			printf("  \"%s\" = \"%s\"\n", key, val);
		}
	}

	printf("\n");

	return 0;
}

static int dumpwtb_dlabel(struct dumpwtb_state* const st, wt_u32 offs, const wtb_dlabelhdr* dlhdr, wt_u32 size)
{
	const unsigned char* last      = ((const unsigned char*)dlhdr) + size - 1;
	const wtb_dlabel*    dlab      = (const wtb_dlabel*)(dlhdr + 1);
	wt_u32               dlabcount = 0; 

	if(size<sizeof(wtb_dlabelhdr))
	{
		printf("DebugLabel section size is less than expected.\n");
		return 1;
	}

	if(*last)
	{
		printf("Last octet of the DebugLabel section is not zero.\n");
		return 1;
	}

	dlabcount = wtr_botoh32(dlhdr->labelcount, st->bo);
	printf("DebugLabel table at offset 0x%X contains %u entries:\n", offs, dlabcount);

	printf("  Addr     Label\n");

	for(;dlabcount;++dlab,--dlabcount)
	{
		wt_u32 addr     = wtr_botoh32(dlab->addr, st->bo);
		wt_u32 lab_offs = wtr_botoh32(dlab->lab_offs, st->bo);

		if(lab_offs>=size)
		{
			printf("DebugLabel label offset is out of range.\n");
			return 1;
		}
		else
		{
			const char* label = ((const char*)dlhdr) + lab_offs;

			printf("  %08X %s\n", addr, label);
		}
	}

	printf("\n");

	return 0;
}

void dumpwtb_add_labels(struct dumpwtb_state* const st, const wtb_dlabelhdr* dlhdr, wt_u32 size)
{
	const wtb_dlabel*    dlab      = (const wtb_dlabel*)(dlhdr + 1);
	wt_u32               dlabcount = wtr_botoh32(dlhdr->labelcount, st->bo);

	for(;dlabcount;++dlab,--dlabcount)
	{
		wt_u32 addr       = wtr_botoh32(dlab->addr, st->bo);
		wt_u32 lab_offs   = wtr_botoh32(dlab->lab_offs, st->bo);
		const char* label = ((const char*)dlhdr) + lab_offs;

		if(lab_offs<size)
		{
			labelmap_add(st->used_labels, addr, label);
		}
	}
}

int dumpwtb(const void* data, wt_u32 size, wt_u32 mode)
{
	struct dumpwtb_state st;
	int                  res = 0;
	char                 buf1[256];
	char                 buf2[256];

	memset(&st, 0, sizeof(st));

	st.data8     = (const wt_u8*)data;
	st.datalen   = size;
	st.hdr       = (const wtb_hdr*)data;
	st.sec_flags = FlagBuilder_Create("");
	st.thr_flags = FlagBuilder_Create("");
	st.disasm    = Disassembler_Create();
	st.labels    = labelmap_create(256);

	if(!st.sec_flags || !st.thr_flags || !st.disasm || !st.labels ||
		FlagBuilder_Add(st.sec_flags, WTBSF_CLEAR, "CLEAR") ||
		FlagBuilder_Add(st.sec_flags, WTBSF_LOAD, "LOAD") ||
		FlagBuilder_Add(st.sec_flags, WTBSF_TYPEVEC, "TYPEVEC") ||
		FlagBuilder_Add(st.thr_flags, WTBTF_EXECUTE, "EXECUTE") ||
		0)
	{
		if(st.sec_flags) { FlagBuilder_Release(st.sec_flags); }
		if(st.thr_flags) { FlagBuilder_Release(st.thr_flags); }
		if(st.disasm) { Disassembler_Release(st.disasm); }
		if(st.labels) { labelmap_delete(st.labels); }
		fprintf(stderr, "Not enough memory to initialize.\n");
		return 2;
	}

	if(st.hdr->b_ident[WTBI_MAGIC0]!=WTBMAG0 ||
		st.hdr->b_ident[WTBI_MAGIC1]!=WTBMAG1 ||
		st.hdr->b_ident[WTBI_MAGIC2]!=WTBMAG2 ||
		st.hdr->b_ident[WTBI_MAGIC3]!=WTBMAG3)
	{
		fprintf(stderr ,"Bad magic number.\n");
		goto return_1;
	}

	if(st.hdr->b_ident[WTBI_VERSION]!=WTBV_CURRENT)
	{
		fprintf(stderr, "Bad binary version.\n");
		goto return_1;
	}

	st.bo = st.hdr->b_ident[WTBI_BYTEORDER];
	if(WTBO_LITTLE!=st.bo && WTBO_BIG!=st.bo)
	{
		fprintf(stderr, "Invalid byteorder.\n");
		goto return_1;
	}

	st.vermatch = WT_FALSE;
	st.bin_version = WT_MAKEVERSION(
		st.hdr->b_ident[WTBI_COMPVER0],
		st.hdr->b_ident[WTBI_COMPVER1],
		st.hdr->b_ident[WTBI_COMPVER2]);
	if(WTR_BINCOMPSINCEVER<=st.bin_version && st.bin_version<=WTR_VERSION)
	{
		st.vermatch = WT_TRUE;
	}

	if(mode & DWM_HEADER)
	{
		wt_u32 ctor = wtr_botoh32(st.hdr->b_ctor, st.bo);
		wt_u32 dtor = wtr_botoh32(st.hdr->b_dtor, st.bo);

		printf("WTB3 Header:\n");
		printf("  Magic:");
		{
			size_t i;
			for(i=0;i<sizeof(st.hdr->b_ident);++i)
			{
				printf(" %02x", st.hdr->b_ident[i]);
			}
			printf("\n");
		}
		printf("  Version:                %u\n", st.hdr->b_ident[WTBI_VERSION]);
		printf("  Byteorder:              %s\n", g_byteorder_names[st.hdr->b_ident[WTBI_BYTEORDER]]);
		printf("  Compiler version:       v%u.%u.%u [%u]\n",
			(unsigned int)WT_GETMAJOR(st.bin_version),
			(unsigned int)WT_GETMINOR(st.bin_version),
			(unsigned int)WT_GETPATCH(st.bin_version),
			(unsigned int)st.hdr->b_ident[WTBI_COMPVER3]);
		printf("  Program size in memory: 0x%X\n", wtr_botoh32(st.hdr->b_size, st.bo));
		printf("  Number of sections:     %u\n", wtr_botoh16(st.hdr->b_shnum, st.bo));
		printf("  Number of threads:      %u\n", wtr_botoh16(st.hdr->b_thnum, st.bo));
		printf("  Sections offset:        0x%X\n", wtr_botoh32(st.hdr->b_shoff, st.bo));
		printf("  Threads offset:         0x%X\n", wtr_botoh32(st.hdr->b_thoff, st.bo));

		printf("  Address of constructor: ");
		if(ctor!=WTR_NULLPC)
		{
			printf("0x%X\n", (wt_u32)(ctor*sizeof(wt_u16)));
		}
		else
		{
			printf("none\n");
		}

		printf("  Address of destructor:  ");
		if(dtor!=WTR_NULLPC)
		{
			printf("0x%X\n", (wt_u32)(dtor*sizeof(wt_u16)));
		}
		else
		{
			printf("none\n");
		}

		printf("\n");
	}

	if(mode & DWM_CODELABELS)
	{
		st.used_labels = st.labels;
	}

	if(mode & DWM_SECTIONS)
	{
		const wtb_shdr* shdr = (const wtb_shdr*)(st.data8 + wtr_botoh32(st.hdr->b_shoff, st.bo));
		wt_u16          i;
		wt_u16          n;

		if(shdr->sh_type!=WTBST_UNUSED || shdr->sh_flag!=0 || shdr->sh_link!=0 || shdr->sh_offs!=0 || shdr->sh_size!=0 || shdr->sh_addr!=0)
		{
			fprintf(stderr, "First section must be empty.");
			goto return_1;
		}

		printf("Section Headers:\n");
		printf("  [Nr] Type       Offs     Size     Addr     Lnk Flags\n");

		for(i=0,n=wtr_botoh16(st.hdr->b_shnum, st.bo);i<n;++shdr,++i)
		{
			wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, st.bo);
			wt_u32 sh_size = wtr_botoh32(shdr->sh_size, st.bo);

			printf("  [%2u] %-10s %08X %08X %08X %3u %X [%s]\n",
				i,
				section_type(shdr->sh_type, buf2, sizeof(buf2)),
				sh_offs,
				sh_size,
				wtr_botoh32(shdr->sh_addr, st.bo),
				wtr_botoh16(shdr->sh_link, st.bo),
				shdr->sh_flag,
				FlagBuilder_Parse(st.sec_flags, shdr->sh_flag, buf1, sizeof(buf1))
				);

			switch(shdr->sh_type)
			{
			case WTBST_DEBUGLABEL:
				if(mode & DWM_CODELABELS)
				{
					dumpwtb_add_labels(&st, (const wtb_dlabelhdr*)(st.data8 + sh_offs), sh_size);
				}
				break;
			}
		}

		printf("\n");
	}
	else
	if(mode & DWM_CODELABELS)
	{
		const wtb_shdr* shdr = (const wtb_shdr*)(st.data8 + wtr_botoh32(st.hdr->b_shoff, st.bo));
		wt_u16          i;
		wt_u16          n;

		for(i=0,n=wtr_botoh16(st.hdr->b_shnum, st.bo);i<n;++shdr,++i)
		{
			if(WTBST_DEBUGLABEL==shdr->sh_type)
			{
				wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, st.bo);
				wt_u32 sh_size = wtr_botoh32(shdr->sh_size, st.bo);

				dumpwtb_add_labels(&st, (const wtb_dlabelhdr*)(st.data8 + sh_offs), sh_size);
			}
		}
	}

	if(mode & DWM_THREADS)
	{
		const wtb_thdr* thdr = (const wtb_thdr*)(st.data8 + wtr_botoh32(st.hdr->b_thoff, st.bo));
		wt_u16          i;
		wt_u16          n;

		printf("Thread Headers:\n");
		printf("  [Nr] Type     CallStk  DataStk  Addr     Flags\n");

		for(i=0,n=wtr_botoh16(st.hdr->b_thnum, st.bo);i<n;++thdr,++i)
		{
			printf("  [%2u] %-8s %08X %08X %08X %X [%s]\n",
				i,
				thread_type(wtr_botoh16(thdr->th_type, st.bo), buf2, sizeof(buf2)),
				wtr_botoh32(thdr->th_cssize, st.bo),
				wtr_botoh32(thdr->th_dssize, st.bo),
				(wt_u32)(wtr_botoh32(thdr->th_pcbase, st.bo) * sizeof(wt_u16)),
				wtr_botoh16(thdr->th_flag, st.bo),
				FlagBuilder_Parse(st.thr_flags, wtr_botoh16(thdr->th_flag, st.bo), buf1, sizeof(buf1))
				);
		}

		printf("\n");
	}

	/* iterate through sections again */

	{
		const wtb_shdr* shdr0 = (const wtb_shdr*)(st.data8 + wtr_botoh32(st.hdr->b_shoff, st.bo));
		const wtb_shdr* shdr  = shdr0;
		wt_u16          i;
		wt_u16          n;

		for(i=0,n=wtr_botoh16(st.hdr->b_shnum, st.bo);i<n;++shdr,++i)
		{
			wt_u16 sh_link = wtr_botoh16(shdr->sh_link, st.bo);
			wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, st.bo);
			wt_u32 sh_size = wtr_botoh32(shdr->sh_size, st.bo);
			wt_u32 sh_addr = wtr_botoh32(shdr->sh_addr, st.bo);

			if(sh_offs)
			{
				switch(shdr->sh_type)
				{
				case WTBST_CODE:
					if(mode & DWM_CODE)
					{
						if(dumpwtb_code(&st, sh_addr, sh_offs, (const wt_u16*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_DATA:
					if(mode & DWM_DATA)
					{
						const wt_u8* bos_data = NULL;
						wt_u32       bos_size = 0;

						if(mode & DWM_DATABOSINTR)
						{
							const wtb_shdr* bos = shdr0 + sh_link;

							if(sh_link>=n)
							{
								fprintf(stderr, "Invalid section index.\n");
								goto return_1;
							}

							bos_data = st.data8 + wtr_botoh32(bos->sh_offs, st.bo);
							bos_size = wtr_botoh32(bos->sh_size, st.bo);
						}

						if(dumpwtb_data(&st, sh_addr, sh_offs, (const wt_u8*)(st.data8 + sh_offs), sh_size, bos_data, bos_size, !!(mode & DWM_ASSUMESTR), !!(shdr->sh_flag & WTBSF_CLEAR))) { goto return_1; }
					}
					break;

				case WTBST_BOS:
					if(mode & DWM_BOS)
					{
						if(dumpwtb_bos(&st, sh_addr, sh_offs, (const wt_u8*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_IMPORT:
					if(mode & DWM_IMPORT)
					{
						if(dumpwtb_import(&st, sh_addr, sh_offs, (const wtb_symhdr*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_EXPORT:
					if(mode & DWM_EXPORT)
					{
						if(dumpwtb_export(&st, sh_addr, sh_offs, (const wtb_symhdr*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_METADATA:
					if(mode & DWM_METADATA)
					{
						if(dumpwtb_metadata(&st, sh_offs, (const wtb_metahdr*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_DEBUGLABEL:
					if(mode & DWM_SHOWDLABELS)
					{
						if(dumpwtb_dlabel(&st, sh_offs, (const wtb_dlabelhdr*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				case WTBST_INTEGRITY:
					if(mode & DWM_INTEGRITY)
					{
						if(dumpwtb_integrity(&st, sh_addr, sh_offs, (const wtb_integhdr*)(st.data8 + sh_offs), sh_size)) { goto return_1; }
					}
					break;

				}
			}
		}
	}

	goto cleanup;

return_1:
	res = 1;
cleanup:
	FlagBuilder_Release(st.sec_flags);
	FlagBuilder_Release(st.thr_flags);
	Disassembler_Release(st.disasm);
	labelmap_delete(st.labels);
	free(st.type_array);

	return res;
}
