#include "config_private.h"
#include "program.h"
#include "type.h"
#include "types.h"
#include "vm.h"
#include "function.h"
#include "funcsig.h"
#include "operator.h"
#include <string.h>

wt_bool wtr_program_resolvetype(struct wtr_program* prog, wt_u32* p_type)
{
	wt_u32 t = *p_type;

	while(WTTF_LOCAL & t)
	{
		wt_u32 idx = t & WTTM_USERTYPEID;

		if(NULL==prog)
		{
			return WT_FALSE;
		}

		if(idx>=prog->num_ref_types)
		{
			return WT_FALSE;
		}

		*p_type = (WTTF_USERTYPE | prog->referenced_types[idx].type->id) | (t & (WTTF_SEQ|WTTF_MUTABLE|WTTF_REF));
		t = *p_type;
	}

	return WT_TRUE;
}

wt_error wtrProgramSetUserPointer(wt_program program, void* ptr)
{
	program->user_ptr = ptr;
	return WTE_OK;
}

wt_error wtrProgramGetUserPointer(wt_program program, void** pptr)
{
	*pptr = program->user_ptr;
	return WTE_OK;
}

const char* wtrProgramGetMetadata(wt_program program, const char* const key)
{
	const size_t num = wtr_vector_size(&program->metadata_vec);

	if(0<num)
	{
		struct wtr_metadata** ptr = (struct wtr_metadata**)
			wtr_vector_at(&program->metadata_vec, 0);

		struct wtr_metadata** const end = ptr + num;

		while(ptr<end)
		{
			struct wtr_metadata* const metadata = *ptr++;

			if(0==strcmp(key, metadata->key))
			{
				return metadata->val;
			}
		}
	}

	return NULL;
}

void wtr_program_free_refd_owned(struct wtr_vm* const vm,
	struct wtr_program* const program, struct wtr_vector* const ownfunc)
{
	size_t i;

	if(0<wtr_vector_size(&program->metadata_vec))
	{
		struct wtr_metadata** ptr = (struct wtr_metadata**)
			wtr_vector_at(&program->metadata_vec, 0);

		struct wtr_metadata** const end = ptr +
			wtr_vector_size(&program->metadata_vec);

		while(ptr<end)
		{
			struct wtr_metadata* const metadata = *ptr++;
			wtr_vm_sys_free(vm, metadata);
		}
	}

	if(0<wtr_vector_size(&program->referenced_functions))
	{
		struct wtr_function** ptr = (struct wtr_function**)
			wtr_vector_at(&program->referenced_functions, 0);

		struct wtr_function** const end = ptr +
			wtr_vector_size(&program->referenced_functions);

		while(ptr<end)
		{
			struct wtr_function* const func = *ptr++;

			if(NULL!=func->program)
			{
				--func->program->refcount;
			}

			--func->refcount;
		}
	}

	if(0<wtr_vector_size(&program->referenced_exceptions))
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&program->referenced_exceptions, 0);

		struct wtr_exception** const end = ptr +
			wtr_vector_size(&program->referenced_exceptions);

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;
			--exc->refcount;
		}
	}

	if(0<wtr_vector_size(&program->referenced_operators))
	{
		struct wtr_operator** ptr = (struct wtr_operator**)
			wtr_vector_at(&program->referenced_operators, 0);

		struct wtr_operator** const end = ptr +
			wtr_vector_size(&program->referenced_operators);

		while(ptr<end)
		{
			struct wtr_operator* const op = *ptr++;
			--op->refcount;
		}
	}

	for(i=0;i<program->num_ref_types;++i)
	{
		--program->referenced_types[i].type->refcount;
	}

	if(0<wtr_vector_size(&program->own_exceptions))
	{
		struct wtr_exception** ptr = (struct wtr_exception**)
			wtr_vector_at(&program->own_exceptions, 0);

		struct wtr_exception** const end = ptr +
			wtr_vector_size(&program->own_exceptions);

		while(ptr<end)
		{
			struct wtr_exception* const exc = *ptr++;
			wtr_vm_removeexception(vm, exc);
		}
	}

	if(0<wtr_vector_size(&program->own_operators))
	{
		struct wtr_operator** ptr = (struct wtr_operator**)
			wtr_vector_at(&program->own_operators, 0);

		struct wtr_operator** const end = ptr +
			wtr_vector_size(&program->own_operators);

		while(ptr<end)
		{
			struct wtr_operator* const op = *ptr++;
			wtr_vm_removeoperator(vm, op);
		}
	}

	if(0<wtr_vector_size(&program->own_types))
	{
		struct wtr_type** ptr = (struct wtr_type**)
			wtr_vector_at(&program->own_types, 0);

		struct wtr_type** const end = ptr +
			wtr_vector_size(&program->own_types);

		while(ptr<end)
		{
			struct wtr_type* const typ = *ptr++;
			wtr_vm_removetype(vm, typ);
		}
	}

	if(NULL==ownfunc)
	{
		struct wtr_list* const list = &vm->list_functions;
		struct wtr_node*       it;

		/* free all functions belong to this program */

		wtr_list_for_each(list, it)
		{
			struct wtr_function* const func =
				wtr_contof(it, struct wtr_function, link_functions);

			if(func->program==program)
			{
				it = wtr_list_prev(it);
				wtr_function_release(vm, func);
			}
		}
	}
	else
	{
		if(0<wtr_vector_size(ownfunc))
		{
			struct wtr_function** ptr = (struct wtr_function**)
				wtr_vector_at(ownfunc, 0);

			struct wtr_function** const end = ptr +
				wtr_vector_size(ownfunc);

			while(ptr<end)
			{
				struct wtr_function* const func = *ptr++;
				wtr_function_release(vm, func);
			}
		}

		wtr_vector_delete(ownfunc, vm);
	}

	wtr_vector_delete(&program->referenced_functions, vm);
	wtr_vector_delete(&program->referenced_exceptions, vm);
	wtr_vector_delete(&program->own_exceptions, vm);
	wtr_vector_delete(&program->referenced_operators, vm);
	wtr_vector_delete(&program->own_operators, vm);
	wtr_vector_delete(&program->own_types, vm);
	wtr_vector_delete(&program->metadata_vec, vm);

	if(NULL!=program->referenced_types)
	{
		wtr_vm_sys_free(vm, program->referenced_types);
	}
}

void wtr_program_destroy(struct wtr_vm* const vm,
	struct wtr_program* const program)
{
	wtr_program_free_refd_owned(vm, program, NULL);

	wtr_list_remove(&program->link_programs);
	wtr_vm_vma_free(vm, program->program_data);

	memset(program, 0, sizeof(struct wtr_program));
	wtr_vm_sys_free(vm, program);
}
